#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import abc
import collections
import collections.abc
import contextlib
import os
import re
import sys

try:
    from scandir import scandir, walk
except ImportError:
    from os import scandir, walk

from functools import wraps
from tempfile import _TemporaryFileWrapper as TemporaryFileWrapperBase

from instakit.utils.misc import memoize, stringify, suffix_searcher, u8bytes, u8str

__all__ = ('DEFAULT_PATH',
           'DEFAULT_PREFIX',
           'DEFAULT_ENCODING',
           'DEFAULT_TIMEOUT',
           'ExecutionError', 'FilesystemError',
           'script_path', 'which', 'back_tick',
           'rm_rf', 'temporary',
           'TemporaryName',
           'Directory',
           'cd', 'wd',
           'TemporaryDirectory', 'Intermediate',
           'NamedTemporaryFile')

__dir__ = lambda: list(__all__)

DEFAULT_PATH = ":".join(filter(os.path.exists, ("/usr/local/bin",
                                                "/bin",  "/usr/bin",
                                                "/sbin", "/usr/sbin")))

DEFAULT_ENCODING = 'latin-1'
DEFAULT_TIMEOUT = 60 # seconds
DEFAULT_PREFIX = "yo-dogg-"

class ExecutionError(Exception):
    """ An error during the execution of a shell command """
    pass

class FilesystemError(Exception):
    """ An error that occurred while mucking about with the filesystem """
    pass

def script_path():
    """ Return the path to the embedded scripts directory. """
    return os.path.join(
           os.path.dirname(__file__), 'scripts')

def which(binary_name, pathvar=None):
    """ Deduces the path corresponding to an executable name,
        as per the UNIX command `which`. Optionally takes an
        override for the $PATH environment variable.
        Always returns a string - an empty one for those
        executables that cannot be found.
    """
    from distutils.spawn import find_executable
    if not hasattr(which, 'pathvar'):
        which.pathvar = os.getenv("PATH", DEFAULT_PATH)
    return find_executable(binary_name, pathvar or which.pathvar) or ""

def back_tick(command,  as_str=True,
                       ret_err=False,
                     raise_err=None, **kwargs):
    """ Run command `command`, return stdout -- or (stdout, stderr) if `ret_err`.
        Roughly equivalent to ``check_output`` in Python 2.7.
        
        Parameters
        ----------
        command : str / list / tuple
            Command to execute. Can be passed as a single string (e.g "ls -la")
            or a tuple or list composed of the commands’ individual tokens (like
            ["ls", "-la"]).
        as_str : bool, optional
            Whether or not the values returned from ``proc.communicate()`` should
            be unicode-decoded as bytestrings (using the specified encoding, which
            defaults to Latin-1) before `back_tick(…)` returns. Default is True.
        ret_err : bool, optional
            If True, the return value is (stdout, stderr). If False, it is stdout.
            In either case `stdout` and `stderr` are strings containing output
            from the commands’ execution. Default is False.
        raise_err : None or bool, optional
            If True, raise instakit.utils.filesystem.errors.ExecutionError when
            calling the function results in a non-zero return code.
            If None, it is set to True if `ret_err` is False,
                                  False if `ret_err` is True.
            Default is None (exception-raising behavior depends on the `ret_err`
            value).
        encoding : str, optional
            The name of the encoding to use when decoding the command output per
            the `as_str` value. Default is “latin-1”.
        directory : str / Directory / path-like, optional
            The directory in which to execute the command. Default is None (in
            which case the process working directory, unchanged, will be used).
        verbose : bool, optional
            Whether or not debug information should be spewed to `sys.stderr`.
            Default is False.
        timeout : int, optional
            Number of seconds to wait for the executed command to complete before
            forcibly killing the subprocess. Default is 60.
        
        Returns
        -------
        out : str / tuple
            If `ret_err` is False, return stripped string containing stdout from
            `command`.  If `ret_err` is True, return tuple of (stdout, stderr)
            where ``stdout`` is the stripped stdout, and ``stderr`` is the
            stripped stderr.
        
        Raises
        ------
        A `instakit.utils.filesystem.errors.ExecutionError` will raise if the
        executed command returns with any non-zero exit status, and `raise_err`
        is set to True.
        
    """
    # Step 1: Prepare for battle:
    import subprocess, shlex
    verbose = bool(kwargs.pop('verbose',  False))
    timeout =  int(kwargs.pop('timeout',  DEFAULT_TIMEOUT))
    encoding = str(kwargs.pop('encoding', DEFAULT_ENCODING))
    raise_err = raise_err is not None and raise_err or bool(not ret_err)
    issequence = isinstance(command, (list, tuple))
    command_str = issequence and " ".join(command) or u8str(command).strip()
    directory = 'directory' in kwargs and os.fspath(kwargs.pop('directory')) or None
    # Step 2: DO IT DOUG:
    if not issequence:
        command = shlex.split(command)
    if verbose:
        print("EXECUTING:", file=sys.stdout)
        print("`{}`".format(command_str),
                            file=sys.stdout)
        print("",           file=sys.stdout)
    process = subprocess.Popen(command, stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                           cwd=directory,
                                         shell=False)
    try:
        output, errors = process.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        process.kill()
        output, errors = process.communicate(timeout=None)
    returncode = process.returncode
    # Step 3: Analyze the return code:
    if returncode is None:
        process.terminate()
        raise ExecutionError('`{}` terminated without exiting cleanly'.format(command_str))
    if raise_err and returncode != 0:
        raise ExecutionError('`{}` exited with status {}, error: “{}”'.format(command_str,
                                   returncode,
                                   u8str(errors).strip()))
    # Step 4: Tidy the output and return it:
    if verbose:
        if returncode != 0:
            print("",                           file=sys.stderr)
            print("NONZERO RETURN STATUS: {}".format(returncode),
                                                file=sys.stderr)
            print("",                           file=sys.stderr)
        if len(u8str(output.strip())) > 0:
            print("")
            print("OUTPUT:",                            file=sys.stdout)
            print("`{}`".format(u8str(output).strip()), file=sys.stdout)
            print("",                                   file=sys.stdout)
        if len(u8str(errors.strip())) > 0:
            print("",                                   file=sys.stderr)
            print("ERRORS:",                            file=sys.stderr)
            print("`{}`".format(u8str(errors).strip()), file=sys.stderr)
            print("",                                   file=sys.stderr)
    output = output.strip()
    if ret_err:
        errors = errors.strip()
        return (as_str and output.decode(encoding) or output), \
               (as_str and errors.decode(encoding) or errors)
    return (as_str and output.decode(encoding) or output)

def rm_rf(pth):
    """ rm_rf() does what `rm -rf` does – so, for the love of fuck,
        BE FUCKING CAREFUL WITH IT.
    """
    if not pth:
        raise ExecutionError(
            "Can’t rm -rf without something to rm arr-effedly")
    pth = os.fspath(pth)
    try:
        if os.path.isfile(pth) or os.path.islink(pth):
            os.unlink(pth)
        elif os.path.isdir(pth):
            subdirs = []
            for path, dirs, files in walk(pth, followlinks=True):
                for tf in files:
                    os.unlink(os.path.join(path, tf))
                subdirs.extend([os.path.join(path, td) for td in dirs])
            for subdir in reversed(subdirs):
                os.rmdir(subdir)
            os.rmdir(pth)
        return True
    except (OSError, IOError):
        pass
    return False

def temporary(suffix=None, prefix=None, parent=None, **kwargs):
    """ Wrapper around `tempfile.mktemp()` that allows full overriding of the
        prefix and suffix by the caller -- that is to say, no random elements
        are used in the returned filename if both a prefix and a suffix are
        supplied.
        
        To avoid problems, the function will throw a FilesystemError if it is
        called with arguments that result in the computation of a filename
        that already exists.
    """
    from tempfile import mktemp, gettempdir
    directory = os.fspath(kwargs.pop('dir', parent) or gettempdir())
    tempmade = mktemp(prefix=prefix, suffix=suffix, dir=directory)
    tempsplit = os.path.splitext(os.path.basename(tempmade))
    if not suffix:
        suffix = tempsplit[1][1:]
    if not prefix or kwargs.pop('randomized', False):
        prefix, _ = os.path.splitext(tempsplit[0]) # WTF, HAX!
    fullpth = os.path.join(directory, "%s%s" % (prefix, suffix))
    if os.path.exists(fullpth):
        raise FilesystemError("temporary(): file exists: %s" % fullpth)
    return fullpth


class TypeLocker(abc.ABCMeta):
    
    """ instakit.utils.filesystem.TypeLocker is a metaclass that does two
        things with the types for whom it is designated as meta:
        
        1) It keeps an index of those types in a dictionary member of
           the `TypeLocker` metaclass itself; and
        
        2) During class creation – the call to `TypeLocker.__new__(…)` –
           it installs a class method called “directory(…)” that will,
           when invoked, always return a new Directory instance that has
           been initialized with the one provided argument “pth” (if one
           was passed).
        
        … The point of this is to allow any of the classes throughout the
        instakit.utils.filesystem module regardless of where they are defined
        or from whom they inherit, to make use of cheaply-constructed Directory
        instances wherever convenient.
        
        Because the “directory(…)” method installed by TypeLocker performs
        a lazy-lookup of the Directory class, using its own type index dict,
        the order of definition does not matter i.e. the TemporaryName class
        (q.v. definition immediately sub.) can use Directories despite its
        definition occuring before Directory – in fact TemporaryName itself
        is utilized within at least one Directory method – sans any issues.
    """
    
    # The metaclass-internal dictionary of generated classes:
    types = collections.OrderedDict()
    
    @classmethod
    def __prepare__(metacls, name, bases, **kwargs):
        """ Maintain declaration order in class members: """
        return collections.OrderedDict()
    
    def __new__(metacls, name, bases, attributes, **kwargs):
        """ All classes are initialized with a “directory(…)”
            class method, lazily returning an instance of the
            instakit.utils.filesystem.Directory(…) class, per
            the arguments:
        """
        # Always replace the “directory” method anew:
        attributes['directory'] = staticmethod(
                                      lambda pth=None: \
                                      metacls.types['Directory'](pth=pth))
        cls = super(TypeLocker, metacls).__new__(metacls, name,
                                                          bases,
                                                          dict(attributes),
                                                        **kwargs)
        metacls.types[name] = cls
        os.PathLike.register(cls)
        return cls

class TemporaryFileWrapper(TemporaryFileWrapperBase,
                           contextlib.AbstractContextManager,
                           os.PathLike,
                           metaclass=TypeLocker):
    
    """ Local subclass of `tempfile._TemporaryFileWrapper`.
        
        We also inherit from both `contextlib.AbstractContextManager`
        and the `os.PathLike` abstract bases -- the latter requires
        that we implement an __fspath__(…) method (q.v. implementation,
        sub.) -- and additionally, `filesystem.TypeLocker` is named as
        the metaclass (q.v. metaclass __new__(…) implementation supra.)
        to cache its type and register it as an os.PathLike subclass.
        
        … Basically a better deal than the original ancestor, like
        all-around. Plus it does not have a name prefixed with an
        underscore, which if it’s not your implementation dogg that
        can be a bit lexically irritating.
    """
    
    def __fspath__(self):
        return self.name

@memoize
def TemporaryNamedFile(tempth, mode='wb', buffer_size=-1, delete=True):
    """ Variation on ``tempfile.NamedTemporaryFile(…)``, for use within
        `filesystem.TemporaryName()` – q.v. class definition sub.
        
        Parameters
        ----------
        tempth : str / bytes / descriptor / filename-ish
            File name, path, or descriptor to open.
        mode : str / bytes, optional
            String-like symbolic explication of mode with which to open
            the file -- q.v. ``io.open(…)`` or ``__builtins__.open(…)``
            supra.
        buffer_size : int, optional
            Integer indicating buffer size to use during file reading
            and/or writing. Default value is -1 (which indicates that
            reads and writes should be unbuffered).
        delete : bool, optional
            Boolean value indicating whether to delete the wrapped
            file upon scope exit or interpreter shutdown (whichever
            happens first). Default is True.
        
        Returns
        -------
            A ``instakit.utils.filesystem.TemporaryFileWrapper`` object,
            initialized and ready to be used, as per its counterpart(s),
            ``tempfile.NamedTemporaryFile``, and
            `filesystem.NamedTemporaryFile`.
        
        Raises
        ------
            A `instakit.utils.filesystem.FilesystemError`, corresponding to
            any errors that may be raised during its own internal calls to
            ``os.open(…)`` and ``os.fdopen(…)``
        
    """
    from tempfile import _bin_openflags, _text_openflags
    
    if 'b' in mode:
        flags = _bin_openflags
    else:
        flags = _text_openflags
    if os.name == 'nt' and delete:
        flags |= os.O_TEMPORARY
    
    descriptor = 0
    filehandle = None
    path = None
    
    try:
        path = os.fspath(tempth)
        descriptor = os.open(path, flags)
        filehandle = os.fdopen(descriptor, mode, buffer_size)
        return TemporaryFileWrapper(filehandle, path, delete)
    except BaseException as base_exception:
        try:
            rm_rf(path)
        except ExecutionError:
            pass
        if descriptor > 0:
            os.close(descriptor)
        raise FilesystemError(str(base_exception))

class TemporaryName(collections.abc.Hashable,
                    contextlib.AbstractContextManager,
                    os.PathLike,
                    metaclass=TypeLocker):
    
    """ This is like NamedTemporaryFile without any of the actual stuff;
        it just makes a file name -- YOU have to make shit happen with it.
        But: should you cause such scatalogical events to transpire, this
        class (when invoked as a context manager) will clean it up for you.
        Unless you say not to. Really it's your call dogg I could give AF
    """
    
    fields = ('name', 'exists',
              'destroy', 'prefix', 'suffix', 'parent')
    
    def __init__(self, prefix=None, suffix="tmp",
                       parent=None,
                     **kwargs):
        """ Initialize a new TemporaryName object.
            
            All parameters are optional; you may specify “prefix”, “suffix”,
            and “dir” (alternatively as “parent” which I think reads better)
            as per `tempfile.mktemp(…)`. Suffixes may omit the leading period
            without confusing things. 
        """
        randomized = kwargs.pop('randomized', False)
        if not prefix:
            prefix = DEFAULT_PREFIX
            randomized = True
        if suffix:
            if not suffix.startswith(os.extsep):
                suffix = "%s%s" % (os.extsep, suffix)
        else:
            suffix = "%stmp" % os.extsep
        if parent is None:
            parent = kwargs.pop('dir', None)
        if parent:
            parent = os.fspath(parent)
        self._name = temporary(prefix=prefix, suffix=suffix,
                                              parent=parent,
                                              randomized=randomized)
        self._destroy = True
        self._parent = parent
        self.prefix = prefix
        self.suffix = suffix
    
    @property
    def name(self):
        """ The temporary file path (which initially does not exist). """
        return self._name
    
    @property
    def basename(self):
        """ The basename (aka the filename) of the temporary file path. """
        return os.path.basename(self._name)
    
    @property
    def dirname(self):
        """ The dirname (aka the enclosing directory) of the temporary file. """
        return self.parent()
    
    @property
    def exists(self):
        """ Whether or not there is anything existant at the temporary file path.
            
            Note that this property will be true for directories created therein,
            as well as FIFOs or /dev entries, or any of the other zany filesystem
            possibilities you and the POSIX standard can imagine, in addition to
            regular files.
        """
        return os.path.exists(self._name)
    
    @property
    def destroy(self):
        """ Whether or not this TemporaryName instance should destroy any file
            that should happen to exist at its temporary file path (as per its
            “name” attribute) on scope exit.
        """
        return self._destroy
    
    @property
    def filehandle(self):
        """ Access a TemporaryNamedFile instance, opened and ready to read and write,
            for this TemporaryName instances’ temporary file path.
            
            Accessing this property delegates the responsibility for destroying
            the TemporaryName file contents to the TemporaryNamedFile object --
            saving the TemporaryNamedFile in, like, a variable somewhere and then
            letting the original TemporaryName go out of scope will keep the file
            alive and unclosed, for example.
        """
        return TemporaryNamedFile(self.do_not_destroy())
    
    def split(self):
        """ Return (dirname, basename) e.g. for /yo/dogg/i/heard/youlike,
            you get back (Directory("/yo/dogg/i/heard"), "youlike")
        """
        return self.dirname, self.basename
    
    def copy(self, destination):
        """ Copy the file (if one exists) at the instances’ file path
            to a new destination.
        """
        if not destination:
            raise FilesystemError("Copying requires a place to which to copy")
        import shutil
        if self.exists:
            return shutil.copy2(self._name, os.fspath(destination))
        return False
    
    def do_not_destroy(self):
        """ Mark this TemporaryName instance as one that should not be automatically
            destroyed upon the scope exit for the instance.
            
            This function returns the temporary file path, and may be called more
            than once without further side effects.
        """
        self._destroy = False
        return self._name
    
    def parent(self):
        """ Sugar for `os.path.abspath(os.path.join(self.name, os.pardir))`
            which, if you are curious, gets you the parent directory of
            the instances’ target filename, wrapped in a Directory
            instance.
        """
        return self.directory(os.path.abspath(
                              os.path.join(self.name,
                                           os.pardir)))
    
    def close(self):
        """ Destroys any existing file at this instances’ file path. """
        if self.exists:
            return rm_rf(self._name)
        return False
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        if self.destroy:
            self.close()
        return exc_type is None
    
    def to_string(self):
        """ Stringify the TemporaryName instance. """
        return stringify(self, type(self).fields)
    
    def __repr__(self):
        return stringify(self, type(self).fields)
    
    def __str__(self):
        if self.exists:
            return os.path.realpath(self._name)
        return self._name
    
    def __bytes__(self):
        return u8bytes(str(self))
    
    def __fspath__(self):
        return self._name
    
    def __bool__(self):
        return self.exists
    
    def __eq__(self, other):
        try:
            return os.path.samefile(self._name,
                                    os.fspath(other))
        except FileNotFoundError:
            return False
    
    def __ne__(self, other):
        try:
            return not os.path.samefile(self._name,
                                        os.fspath(other))
        except FileNotFoundError:
            return True
    
    def __hash__(self):
        return hash((self._name, self.exists))

non_dotfile_match = re.compile(r"^[^\.]").match
non_dotfile_matcher = lambda p: non_dotfile_match(p.name) # type: ignore

class Directory(collections.abc.Hashable,
                collections.abc.Mapping,
                collections.abc.Sized,
                contextlib.AbstractContextManager,
                os.PathLike,
                metaclass=TypeLocker):
    
    """ A context-managed directory: change in on enter, change back out
        on exit. Plus a few convenience functions for listing and whatnot.
    """
    
    fields = ('name', 'old', 'new', 'exists',
              'will_change',        'did_change',
              'will_change_back',   'did_change_back')
    
    zip_suffix = "%szip" % os.extsep
    
    def __init__(self, pth=None):
        """ Initialize a new Directory object.
            
            There is only one parameter, “pth” -- the target path for the Directory
            object instance. When the Directory is initialized as a context manager,
            the process working directory will change to this directory (provided
            that it’s a different path than the current working directory, according
            to `os.path.samefile(…)`).
            
            The “pth” parameter is optional, in which case the instance uses the
            process working directory as its target, and no change-of-directory
            calls will be issued. Values for “pth” can be string-like, or existing
            Directory instances -- either will work.
            
            There are two decendant classes of Directory (q.v. definitions below)
            that enforce stipulations for the “pth” parameter: the `cd` class 
            requires a target path to be provided (and therefore will nearly always
            change the working directory when invoked as a context manager). Its
            sibling class `wd` forbids the naming of a “pth” value, thereby always
            initializing itself with the current working directory as its target,
            and fundamentally avoids issuing any directory-change calls.
        """
        if pth is not None:
            self.target = u8str(os.fspath(pth))
        else:
            self.target = os.getcwd()
    
    @property
    def name(self):
        """ The instances’ target directory path. """
        return getattr(self, 'target', None) or \
               getattr(self, 'new')
    
    @property
    def basename(self):
        """ The basename (aka the name of the directory, like as opposed to the
            entire fucking absolute path) of the target directory.
        """
        return os.path.basename(self.name)
    
    @property
    def dirname(self):
        """ The dirname (aka the path of the enclosing directory) of the target
            directory, wrapped in a new Directory instance.
        """
        return self.parent()
    
    @property
    def exists(self):
        """ Whether or not the instances’ target path exists as a directory. """
        return os.path.isdir(self.name)
    
    @property
    def initialized(self):
        """ Whether or not the instance has been “initialized” -- as in, the
            `target` instance value has been set (q.v. `ctx_initialize(…)`
            help sub.) as it stands immediately after `__init__(…)` has run.
        """
        return hasattr(self, 'target')
    
    @property
    def targets_set(self):
        """ Whether or not the instance has had targets set (the `new` and `old`
            instance values, q.v. `ctx_set_targets(…)` help sub.) and is ready
            for context-managed use.
        """
        return hasattr(self, 'old') and hasattr(self, 'new')
    
    @property
    def prepared(self):
        """ Whether or not the instance has been internally prepared for use
            (q.v. `ctx_prepare()` help sub.) and is in a valid state.
        """
        return hasattr(self, 'will_change') and \
               hasattr(self, 'will_change_back') and \
               hasattr(self, 'did_change') and \
               hasattr(self, 'did_change_back')
    
    def split(self):
        """ Return a two-tuple containing `(dirname, basename)` – like e.g.
            for `/yo/dogg/i/heard/youlike`, your return value will be like
            `(Directory("/yo/dogg/i/heard"), "youlike")`
        """
        return self.dirname, self.basename
    
    def ctx_initialize(self):
        """ Restores the instance to the freshly-allocated state -- with one
            notable exception: if it had been previously prepared (through a
            call to `instance.ctx_prepare()`) and thus has a “new” attribute
            filled in with a target path, `ctx_initialize()` will preserve
            the contents of that attribute in the value of the `self.target` 
            instance member.
            
            The call deletes all other instance attributes from the internal
            mapping of the instance in question, leaving it in a state ready
            for either context-managed reëntry, or for reuse in an unmanaged
            fashion *provided* one firstly calls `instance.ctx_set_targets()`
            or `instance.ctx_prepare()` in order to reconfigure (the minimal
            subset of, or the full complement of) the member-variable values
            needed by the internal workings of a Directory instance.
        """
        if self.targets_set:
            self.target = u8str(self.new or self.old)
            del self.old
            del self.new
        if self.prepared:
            del self.will_change
            del self.will_change_back
            del self.did_change
            del self.did_change_back
        return self
    
    def ctx_set_targets(self, old=None):
        """ Sets the “self.old” and “self.new” instance variable values,
            using the value of `self.target` and an (optional) string-like
            argument to use as the value for “self.old”.
            
            One shouldn’t generally call this or have a need to call this --
            although one can manually invoke `instance.ctx_set_targets(…)`
            to reconfigure a Directory instance to use it again after it has
            been re-initialized after a call to `instance.ctx_initialize()`
            (q.v. `ctx_initialize()` help supra.) in cases where it isn’t
            going to be used as part of a managed context; that is to say,
            outside of a `with` statement.
            
            (Within a `with` statement, the call issued upon scope entry to
            `Directory.__enter__(self)` will internally make a call to
            `Directory.ctx_prepare(self)` (q.v. doctext help sub.) which
            that will call `Directory.ctx_set_targets(self, …)` itself.)
        """
        if not self.initialized:
            if old is None:
                old = os.getcwd()
            setattr(self, 'old', old)
            setattr(self, 'new', old)
            return self
        setattr(self, 'old', old is not None and old or self.target)
        setattr(self, 'new', self.target)
        del self.target
        return self
    
    def ctx_prepare(self):
        """ Prepares the member values of the Directory instance according
            to a requisite `self.target` directory-path value; the primary
            logic performed by this function determines whether or not it
            is necessary to switch the process working directory while the
            Directory instance is actively being used as a context manager
            in the scope of a `while` block.
            
            The reason this is done herein is to minimize the number of
            calls to potentially expensive system-call-wrapping functions
            such as `os.getcwd()`, `os.path.samefile(…)`, and especially
            `os.chdir(…)` -- which the use of the latter affects the state
            of the process issuing the call in a global fashion, and can
            cause invincibly undebuggable behavioral oddities to crop up
            in a variety of circumstances. 
        """
        self.ctx_set_targets(old=os.getcwd())
        if os.path.isdir(self.new):
            self.will_change = not os.path.samefile(self.old,
                                                    self.new)
        else:
            self.will_change = False
        self.did_change = False
        self.will_change_back = self.will_change
        self.did_change_back = False
        return self
    
    def __enter__(self):
        self.ctx_prepare()
        if self.will_change and self.exists:
            os.chdir(self.new)
            self.did_change = os.path.samefile(self.new,
                                               os.getcwd())
            self.will_change_back = self.did_change
        return self
    
    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        # N.B. return False to throw, True to supress:
        if self.will_change_back and os.path.isdir(self.old):
            os.chdir(self.old)
            self.did_change_back = os.path.samefile(self.old,
                                                    os.getcwd())
            if self.did_change_back:
                # return to pristine state:
                self.ctx_initialize()
                return exc_type is None
        # return to pristine state:
        self.ctx_initialize()
        return False
    
    def realpath(self, pth=None):
        """ Sugar for calling os.path.realpath(self.name) """
        return u8str(
            os.path.realpath(
            os.fspath(pth or self.name)))
    
    def ls(self, pth=None, suffix=None):
        """ List files -- defaults to the process’ current working directory.
            As per the UNIX custom, files whose name begins with a dot are
            omitted.
            
            Specify an optional “suffix” parameter to filter the list by a
            particular file suffix (leading dots unnecessary but unharmful).
        """
        files = (str(direntry.name) \
                 for direntry in filter(non_dotfile_matcher,
                                scandir(self.realpath(pth))))
        if not suffix:
            return files
        return filter(suffix_searcher(suffix), files)
    
    def ls_la(self, pth=None, suffix=None):
        """ List all files, including files whose name starts with a dot.
            The default is to use the process’ current working directory.
            
            Specify an optional “suffix” parameter to filter the list by a
            particular file suffix (leading dots unnecessary but unharmful).
            
            (Technically speaking, `ls_la()` is a misnomer for this method,
            as it does not provide any extended meta-info like you get if
            you use the “-l” flag when invoking the `ls` command -- I just
            like calling it that because “ls -la” was one of the first shell
            commands I ever learned, and it reads better than `ls_a()` which
            I think looks awkward and goofy.)
        """
        files = (str(direntry.name) \
                 for direntry in scandir(self.realpath(pth)))
        if not suffix:
            return files
        return filter(suffix_searcher(suffix), files)
    
    def subpath(self, subpth, whence=None, requisite=False):
        """ Returns the path to a subpath of the instances’ target path. """
        fullpth = os.path.join(os.fspath(whence or self.name),
                               os.fspath(subpth))
        return (os.path.exists(fullpth) or not requisite) and fullpth or None
    
    def subdirectory(self, subdir, whence=None):
        """ Returns the path to a subpath of the instances’ target path --
            much like Directory.subpath(…) -- as an instance of Directory.
        """
        pth = self.subpath(subdir, whence, requisite=False)
        if os.path.isfile(pth):
            raise FilesystemError("file exists at subdirectory path: %s" % pth)
        if os.path.islink(pth):
            raise FilesystemError("symlink exists at subdirectory path: %s" % pth)
        if os.path.ismount(pth):
            raise FilesystemError("mountpoint exists at subdirectory path: %s" % pth)
        return self.directory(pth)
    
    def makedirs(self, pth=None):
        """ Creates any parts of the target directory path that don’t
            already exist, á la the `mkdir -p` shell command.
        """
        try:
            os.makedirs(os.path.abspath(
                        os.path.join(self.name,
                        os.fspath(pth or os.curdir))),
                        exist_ok=False)
        except OSError as os_error:
            raise FilesystemError(str(os_error))
        return self
    
    def walk(self, followlinks=False):
        """ Sugar for calling X.walk(self.name), where X is either
            `scandir` (in the case of python 2.7) or `os` (for
            python 3 and thereafter).
        """
        return walk(self.name, followlinks=followlinks)
    
    def parent(self):
        """ Sugar for `os.path.abspath(os.path.join(self.name, os.pardir))`
            which, if you are curious, gets you the parent directory of
            the instances’ target directory, wrapped in a Directory
            instance.
        """
        return self.directory(os.path.abspath(
                              os.path.join(self.name,
                                           os.pardir)))
    
    def copy_all(self, destination):
        """ Copy the entire directory tree, all contents included, to a new
            destination path. The destination must not already exist, and
           `copy_all(…)` will not overwrite existant directories. Like, if
            you have yourself an instance of Directory, `directory`, and you
            want to copy it to `/home/me/myshit`, `/home/me` should already
            exist but `/home/me/myshit` should not, as the subdirectory
           `myshit` gets created by the `directory.copy_all('/home/me/myshit')`
            invocation (like as a side-effect).
            
            Does that make sense to you? Try it, you’ll get a `FilesystemError`
            if it evidently did not make sense to you.
            
            The destination path may be specified using a string-like, with
            another Directory object, or anything deemed path-y enough by
           `os.fspath(…)`. Internally, this method uses `shutil.copytree(…)`
            to tell the filesystem what to copy where.
        """
        import shutil
        whereto = self.directory(pth=destination)
        if whereto.exists or os.path.isfile(whereto.name) \
                          or os.path.islink(whereto.name):
            raise FilesystemError(
                "copy_all() destination exists: %s" % whereto.name)
        if self.exists:
            return shutil.copytree(self.name, whereto.name)
        return False
    
    def zip_archive(self, zpth=None, zmode=None):
        """ Recursively descends through the target directory, stowing all
            that it finds into a zipfile at the specified path.
            
            Use the optional “zmode” parameter to specify the compression
            algorithm, as per the constants found in the `zipfile` module;
            the default value is `zipfile.ZIP_DEFLATED`.
        """
        import zipfile
        if zpth is None:
            raise FilesystemError("Need to specify a zip-archive file path")
        zpth = os.fspath(zpth)
        if not zpth.lower().endswith(self.zip_suffix):
            zpth += self.zip_suffix
        if os.path.exists(zpth):
            if os.path.isdir(zpth):
                raise FilesystemError("Can't overwrite a directory: %s" % zpth)
            raise FilesystemError("File path for zip-archive already exists")
        if not zmode:
            zmode = zipfile.ZIP_DEFLATED
        with TemporaryName(prefix="ziparchive-",
                           suffix=self.zip_suffix[1:]) as ztmp:
            with zipfile.ZipFile(ztmp.name, "w", zmode) as ziphandle:
                relparent = lambda p: os.path.relpath(p, os.fspath(self.parent()))
                for root, dirs, files in self.walk(followlinks=True):
                    ziphandle.write(root, relparent(root)) # add directory
                    for filename in files:
                        filepath = os.path.join(root, filename)
                        if os.path.isfile(filepath): # regular files only
                            arcname = os.path.join(relparent(root), filename)
                            ziphandle.write(filepath, arcname) # add regular file
            ztmp.copy(zpth)
        return self.realpath(zpth)
    
    def close(self):
        """ Stub method -- always returns True: """
        return True
    
    def to_string(self):
        """ Stringify the Directory instance. """
        return stringify(self, type(self).fields)
    
    def __repr__(self):
        return stringify(self, type(self).fields)
    
    def __str__(self):
        if self.exists:
            return self.realpath()
        return self.name
    
    def __bytes__(self):
        return u8bytes(str(self))
    
    def __fspath__(self):
        return self.name
    
    def __bool__(self):
        return self.exists
    
    def __iter__(self):
        return scandir(self.realpath())
    
    def __len__(self):
        return len(list(self))
    
    def __getitem__(self, filename):
        pth = self.subpath(filename, requisite=True)
        if not pth:
            raise KeyError(
                "file not found: %s" % os.fspath(filename))
        return pth
    
    def __contains__(self, filename):
        return self.subpath(filename, requisite=True) is not None
    
    def __eq__(self, other):
        try:
            return os.path.samefile(self.name,
                                    os.fspath(other))
        except FileNotFoundError:
            return False
    
    def __ne__(self, other):
        try:
            return not os.path.samefile(self.name,
                                        os.fspath(other))
        except FileNotFoundError:
            return True
    
    def __hash__(self):
        return hash((self.name, self.exists))
    
    def __call__(self, *args, **kwargs):
        return self

class cd(Directory):
    
    def __init__(self, pth):
        """ Change to a new directory (the target path `pth` must be specified).
        """
        super(cd, self).__init__(pth)


class wd(Directory):
    
    def __init__(self):
        """ Initialize a Directory instance for the current working directory.
        """
        super(wd, self).__init__(pth=None)


class TemporaryDirectory(Directory):
    
    """ It's funny how this code looks, like, 99 percent exactly like the above
        TemporaryName class -- shit just works out that way. But this actually
        creates the directory in question; much like filesystem::TemporaryDirectory
        from libimread, this class wraps tempfile.mkdtemp() and can be used as
        a context manager (the C++ orig used RAII).
    """
    
    fields = ('name', 'old', 'new', 'exists',
              'destroy', 'prefix',  'suffix', 'parent',
              'will_change',        'did_change',
              'will_change_back',   'did_change_back')
    
    def __init__(self, prefix="TemporaryDirectory-", suffix="",
                                                     parent=None,
                                                     change=True,
                                                   **kwargs):
        """ Initialize a new TemporaryDirectory object.
            
            All parameters are optional; you may specify “prefix”, “suffix”,
            and “dir” (alternatively as “parent” which I think reads better)
            as per `tempfile.mkdtemp(…)`. Suffixes may omit the leading period
            without confusing things. 
            
            The boolean “change” parameter determines whether or not the
            process working directory will be changed to the newly created
            temporary directory; it defaults to `True`.
        """
        from tempfile import mkdtemp
        if suffix:
            if not suffix.startswith(os.extsep):
                suffix = "%s%s" % (os.extsep, suffix)
        if parent is None:
            parent = kwargs.pop('dir', None)
        if parent:
            parent = os.fspath(parent)
        self._name = mkdtemp(prefix=prefix, suffix=suffix, dir=parent)
        self._destroy = True
        self._parent = parent
        self.prefix = prefix
        self.suffix = suffix
        self.change = change
        super(TemporaryDirectory, self).__init__(self._name)
    
    @property
    def name(self):
        """ The temporary directory pathname. """
        return self._name
    
    @property
    def exists(self):
        """ Whether or not the temporary directory exists. """
        return os.path.isdir(self._name)
    
    @property
    def destroy(self):
        """ Whether or not the TemporaryDirectory instance has been marked
            for automatic deletion upon scope exit (q.v __exit__(…) method
            definition sub.)
        """
        return self._destroy
    
    @wraps(Directory.ctx_prepare)
    def ctx_prepare(self):
        change = super(TemporaryDirectory, self).ctx_prepare().change
        self.will_change = self.will_change_back = bool(self.will_change and change)
        return self
    
    def close(self):
        """ Delete the directory pointed to by the TemporaryDirectory
            instance, and everything it contains. USE WITH CAUTION.
        """
        out = super(TemporaryDirectory, self).close()
        if self.exists:
            return rm_rf(self.name) and out
        return False
    
    def do_not_destroy(self):
        """ Mark this TemporaryDirectory instance as one that should not
            be automatically destroyed upon the scope exit for the instance.
            
            This function returns the temporary directory path, and may
            be called more than once without further side effects.
        """
        self._destroy = False
        return self.name
    
    def __enter__(self):
        if not self.exists:
            self.makedirs()
        super(TemporaryDirectory, self).__enter__()
        return self
    
    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        out = super(TemporaryDirectory, self).__exit__(exc_type, exc_val, exc_tb)
        if self.destroy:
            out &= self.close()
        return out


class Intermediate(TemporaryDirectory, Directory):
    
    """ instakit.utils.filesystem.Intermediate isn’t a class, per se – rather,
        it is a class factory proxy that normally constructs a new Directory
        instance in leu of itself, except for when it it is constructed without
        a `pth` argument, in which case, it falls back to the construction of
        a new TemporaryDirectory instance instead.
    """
    
    def __new__(cls, pth=None):
        """ The constructor simply delegates to the creation of either a new
            Directory or a new TemporaryDirectory.
        """
        if pth is not None:
            return Directory(pth=pth)
        return TemporaryDirectory(prefix="%s-" % cls.__name__,
                                  change=False)
    
    def __init__(self, pth=None):
        """ The initializer explicitly does nothing, as it will always be called
            on an already-initialized instance of some other class.
        """
        pass

def NamedTemporaryFile(mode='w+b', buffer_size=-1,
                       suffix="tmp", prefix=DEFAULT_PREFIX,
                       directory=None,
                       delete=True):
    """ Variation on tempfile.NamedTemporaryFile(…), such that suffixes
        are passed WITHOUT specifying the period in front (versus the
        standard library version which makes you pass suffixes WITH
        the fucking period, ugh).
    """
    from tempfile import (gettempdir, _bin_openflags,
                                      _text_openflags,
                                      _mkstemp_inner)
    
    parent = Directory(pth=directory or gettempdir())
    
    if suffix:
        if not suffix.startswith(os.extsep):
            suffix = "%s%s" % (os.extsep, suffix)
    else:
        suffix = "%stmp" % os.extsep
    
    if 'b' in mode:
        flags = _bin_openflags
    else:
        flags = _text_openflags
    if os.name == 'nt' and delete:
        flags |= os.O_TEMPORARY
    
    (descriptor, name) = _mkstemp_inner(parent.name, prefix,
                                                     suffix, flags,
                                             u8bytes(suffix))
    try:
        filehandle = os.fdopen(descriptor, mode, buffer_size)
        return TemporaryFileWrapper(filehandle, name, delete)
    except BaseException as base_exception:
        rm_rf(name)
        if descriptor > 0:
            os.close(descriptor)
        raise FilesystemError(str(base_exception))

del TemporaryFileWrapperBase

def test():
    
    """ Run the inline tests for the instakit.utils.filesystem module. """
    
    # Simple inline tests for “TemporaryName”, “cd” and “cwd”,
    # and “TemporaryDirectory”:
    from tempfile import gettempdir
    
    initial = os.getcwd()
    tfp = None
    tdp = None
    
    with TemporaryName(prefix="test-temporaryname-",
                       randomized=True) as tfn:
        print("* Testing TemporaryName file instance: %s" % tfn.name)
        assert os.path.samefile(os.getcwd(),            initial)
        assert gettempdir() in tfn.name
        assert tfn.prefix == "test-temporaryname-"
        assert tfn.suffix == ".tmp"
        assert not tfn._parent
        assert tfn.prefix in tfn.name
        assert tfn.suffix in tfn.name
        assert tfn.prefix in os.fspath(tfn)
        assert tfn.suffix in os.fspath(tfn)
        assert tfn.destroy
        assert type(tfn.directory(os.path.basename(tfn.name))) == Directory
        assert isinstance(tfn,                          TemporaryName)
        assert isinstance(tfn,                          collections.abc.Hashable)
        assert isinstance(tfn,                          contextlib.AbstractContextManager)
        assert isinstance(tfn,                          os.PathLike)
        p = tfn.parent()
        assert os.path.samefile(os.fspath(p),           gettempdir())
        # The next four asserts will be “not” asserts while
        # the TemporaryName has not been written to:
        assert not tfn.exists
        assert not os.path.isfile(os.fspath(tfn))
        assert not tfn.basename in tfn.dirname
        assert not tfn.basename in p
        # Here we write something to the TemporaryName:
        with open(os.fspath(tfn), mode="w") as handle:
            handle.write("yo dogg")
        # Now we repeat the above four asserts,
        # as positive assertions:
        assert tfn.exists
        assert os.path.isfile(os.fspath(tfn))
        assert tfn.basename in tfn.dirname
        assert tfn.basename in p
        # Stash the TemporaryName’s path to later assert
        # that it no longer exists - that it has been correctly
        # deleted on scope exit:
        tfp = tfn.name
        assert os.path.exists(tfp)
        print("* TemporaryName file object tests completed OK")
        print("")
    
    # Confirm that the TemporaryName has been deleted:
    assert not os.path.exists(tfp)
    
    with wd() as cwd:
        print("* Testing working-directory instance: %s" % cwd.name)
        assert os.path.samefile(os.getcwd(),           cwd.new)
        assert os.path.samefile(os.getcwd(),           cwd.old)
        assert os.path.samefile(os.getcwd(),           os.fspath(cwd))
        assert os.path.samefile(cwd.new,               cwd.old)
        assert os.path.samefile(cwd.new,               initial)
        assert os.path.samefile(cwd.old,               initial)
        assert os.path.samefile(cwd.new,               os.fspath(cwd))
        assert os.path.samefile(cwd.old,               os.fspath(cwd))
        assert os.path.samefile(os.fspath(cwd),        initial)
        assert not cwd.subdirectory('yodogg').exists
        # assert cwd.subdirectory('yodogg').makedirs().exists
        assert not cwd.will_change
        assert not cwd.did_change
        assert not cwd.will_change_back
        assert not cwd.did_change_back
        assert type(cwd.directory(cwd.new)) == Directory
        assert isinstance(cwd,                         wd)
        assert isinstance(cwd,                         Directory)
        assert isinstance(cwd,                         collections.abc.Hashable)
        assert isinstance(cwd,                         collections.abc.Mapping)
        assert isinstance(cwd,                         collections.abc.Sized)
        assert isinstance(cwd,                         contextlib.AbstractContextManager)
        assert isinstance(cwd,                         os.PathLike)
        assert os.path.isdir(os.fspath(cwd))
        assert not 'yodogg' in cwd
        assert cwd.basename in cwd.dirname
        # print(", ".join(list(cwd.ls())))
        print("* Working-directory object tests completed OK")
        print("")
    
    with cd(gettempdir()) as tmp:
        print("* Testing directory-change instance: %s" % tmp.name)
        assert os.path.samefile(os.getcwd(),          gettempdir())
        assert os.path.samefile(os.getcwd(),          tmp.new)
        assert os.path.samefile(gettempdir(),         tmp.new)
        assert os.path.samefile(os.getcwd(),          os.fspath(tmp))
        assert os.path.samefile(gettempdir(),         os.fspath(tmp))
        assert not os.path.samefile(os.getcwd(),      initial)
        assert not os.path.samefile(tmp.new,          initial)
        assert not os.path.samefile(os.fspath(tmp),   initial)
        assert os.path.samefile(tmp.old,              initial)
        # assert not tmp.subdirectory('yodogg').exists
        # assert tmp.subdirectory('yodogg').makedirs().exists
        assert tmp.will_change
        assert tmp.did_change
        assert tmp.will_change_back
        assert not tmp.did_change_back
        assert type(tmp.directory(tmp.new)) == Directory
        assert isinstance(tmp,                        cd)
        assert isinstance(tmp,                        Directory)
        assert isinstance(tmp,                        collections.abc.Hashable)
        assert isinstance(tmp,                        collections.abc.Mapping)
        assert isinstance(tmp,                        collections.abc.Sized)
        assert isinstance(tmp,                        contextlib.AbstractContextManager)
        assert isinstance(tmp,                        os.PathLike)
        assert os.path.isdir(os.fspath(tmp))
        assert tmp.basename in tmp.dirname
        print("* Directory-change object tests completed OK")
        print("")
    
    with TemporaryDirectory(prefix="test-temporarydirectory-") as ttd:
        print("* Testing TemporaryDirectory instance: %s" % ttd.name)
        # assert os.path.commonpath((os.getcwd(), gettempdir())) == gettempdir()
        # print(os.path.commonpath((os.getcwd(), gettempdir())))
        assert gettempdir() in ttd.name
        assert gettempdir() in ttd.new
        assert gettempdir() in os.fspath(ttd)
        assert initial not in ttd.name
        assert initial not in ttd.new
        assert initial not in os.fspath(ttd)
        assert initial in ttd.old
        assert not ttd.subdirectory('yodogg').exists
        assert ttd.subdirectory('yodogg').makedirs().exists
        assert 'yodogg' in ttd
        assert ttd.prefix == "test-temporarydirectory-"
        assert not ttd.suffix
        assert not ttd._parent
        assert ttd.prefix in ttd.name
        assert ttd.exists
        assert ttd.destroy
        assert ttd.will_change
        assert ttd.did_change
        assert ttd.will_change_back
        assert not ttd.did_change_back
        assert type(ttd.directory(ttd.new)) == Directory
        assert isinstance(ttd,                          TemporaryDirectory)
        assert isinstance(ttd,                          Directory)
        assert isinstance(ttd,                          collections.abc.Hashable)
        assert isinstance(ttd,                          collections.abc.Mapping)
        assert isinstance(ttd,                          collections.abc.Sized)
        assert isinstance(ttd,                          contextlib.AbstractContextManager)
        assert isinstance(ttd,                          os.PathLike)
        p = ttd.parent()
        assert os.path.samefile(os.fspath(p),           gettempdir())
        assert os.path.isdir(os.fspath(ttd))
        assert os.path.isdir(os.fspath(p))
        assert ttd.basename in p
        assert ttd.basename in ttd.dirname
        assert ttd.dirname == p
        # Stash the TemporaryDirectory’s path as a Directory
        # instance, to later assert that it no longer exists –
        # that it has been correctly deleted on scope exit:
        tdp = Directory(ttd)
        assert tdp.exists
        print("* TemporaryDirectory object tests completed OK")
        print("")
    
    # Confirm that the TemporaryDirectory has been deleted:
    assert not tdp.exists

if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
from __future__ import print_function

import os

from clu.exporting import ExporterBase

# The “prefix” is the directory enclosing the package root:
prefix = os.path.dirname(
         os.path.dirname(__file__))

class Exporter(ExporterBase, prefix=prefix, appname="instakit"):
    pass

exporter = Exporter(path=__file__)
export = exporter.decorator()

export(Exporter)

# Assign the modules’ `__all__` and `__dir__` using the exporter:
__all__, __dir__ = exporter.all_and_dir()



#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False

from cython.operator cimport dereference as deref
from cpython.float cimport PyFloat_AS_DOUBLE
from libcpp.memory cimport unique_ptr

from instakit.comparators.ext.butteraugli cimport Image8, ImageF
from instakit.comparators.ext.butteraugli cimport image8vec, imagefvec
from instakit.comparators.ext.butteraugli cimport floatvecvec, CreatePlanes
from instakit.utils.mode import Mode

ctypedef unique_ptr[ImageF] imagef_ptr
ctypedef unique_ptr[Image8] image8_ptr

cdef imagefvec image_to_planar_vector(object pilimage):
    cdef int width, height, x, y
    cdef double point
    cdef float* planerow
    cdef object pypoint, image, accessor
    cdef object bands, band
    cdef imagef_ptr plane
    cdef imagefvec planes
    cdef int bandcount, idx
    
    width, height = pilimage.size
    bands = Mode.RGB.process(pilimage).split()
    
    with nogil:
        bandcount = 3
        planes = CreatePlanes[float](width,
                                     height,
                                     bandcount)
    
    for idx in range(bandcount):
        band = bands[idx]
        image = Mode.F.process(band)
        accessor = image.load()
        
        for y in range(height):
            planerow = planes[idx].Row(y)
            for x in range(width):
                pypoint = accessor[x, y]
                point = PyFloat_AS_DOUBLE(pypoint)
                planerow[x] = <float>point
    
    # AND SO NOW WHAT??!



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#    
#    INSTAKIT -- Instagrammy image-processors and tools, based on Pillow and SciPy
#    
#    Copyright © 2012-2025 Alexander Böhn
#    
#    Permission is hereby granted, free of charge, to any person obtaining a copy 
#    of this software and associated documentation files (the "Software"), to deal 
#    in the Software without restriction, including without limitation the rights 
#    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
#    copies of the Software, and to permit persons to whom the Software is 
#    furnished to do so, subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be included in all 
#    copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
#    SOFTWARE.
#
''' Image processors for django-imagekit - based on Pillow, SciPy, and scikit-image '''

from __future__ import print_function
import os, sys, sysconfig

try:
    from os import cpu_count
except ImportError:
    try:
        from multiprocessing import cpu_count
    except ImportError:
        try:
            from psutil import cpu_count
        except ImportError:
            cpu_count = lambda: 1

from setuptools import setup, find_packages
from Cython.Build import cythonize

# HOST PYTHON VERSION
PYTHON_VERSION = float("%s%s%s" % (sys.version_info.major, os.extsep,
                                   sys.version_info.minor))

# CONSTANTS
PROJECT_NAME = 'instakit'
AUTHOR_NAME = 'Alexander Böhn'
AUTHOR_USER = 'fish2000'

GITHUB = 'github.com'
GMAIL = 'gmail.com'

AUTHOR_EMAIL = '%s@%s' % (AUTHOR_USER, GMAIL)
PROJECT_GH_URL = 'https://%s/%s/%s' % (GITHUB,
                                       AUTHOR_USER,
                                       PROJECT_NAME)
PROJECT_DL_URL = '%s/zipball/master' % PROJECT_GH_URL

KEYWORDS = ('django',
            'imagekit', PROJECT_NAME,
                        AUTHOR_USER,
            'image processing',
            'image analysis',
            'image comparison',
            'halftone', 'dithering',
            'Photoshop', 'acv', 'curves',
            'PIL', 'Pillow',
            'Cython',
            'NumPy', 'SciPy', 'scikit-image')

CPPLANGS = ('c++', 'cxx', 'cpp', 'cc', 'mm')
CPPVERSION = PYTHON_VERSION < 3 and 'c++14' or 'c++17'

# PROJECT DIRECTORY
CWD = os.path.dirname(__file__)
BASE_PATH = os.path.join(
            os.path.abspath(CWD), PROJECT_NAME)

def project_content(*filenames):
    import io
    filepath = os.path.join(CWD, *filenames)
    if not os.path.isfile(filepath):
        raise IOError("""File %s doesn't exist""" % filepath)
    out = ''
    with io.open(filepath, 'r') as handle:
        out += handle.read()
    if not out:
        raise ValueError("""File %s couldn't be read""" % os.path.sep.join(filenames))
    return out.strip()

# CYTHON & C-API EXTENSION MODULES
def cython_module(*args, **kwargs):
    from Cython.Distutils import Extension
    sources = []
    sources.extend(kwargs.pop('sources', []))
    include_dirs = []
    include_dirs.extend(kwargs.pop('include_dirs', []))
    ext_package = os.path.extsep.join(args)
    ext_pth = os.path.sep.join(args) + os.extsep + "pyx"
    sources.insert(0, ext_pth)
    language = kwargs.pop('language', 'c').lower()
    extra_compile_args = ['-Wno-unused-function',
                          '-Wno-unneeded-internal-declaration',
                          '-O3',
                          '-fstrict-aliasing',
                          '-funroll-loops',
                          '-mtune=native']
    if language in CPPLANGS:
        extra_compile_args.extend(['-std=%s' % CPPVERSION,
                                   '-stdlib=libc++',
                                   '-Wno-sign-compare',
                                   '-Wno-unused-private-field'])
    return Extension(ext_package, sources,
        language=language,
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args)

def cython_comparator(name, **kwargs):
    return cython_module(PROJECT_NAME, 'comparators', 'ext', name, **kwargs)

def cython_processor(name, **kwargs):
    return cython_module(PROJECT_NAME, 'processors', 'ext', name, **kwargs)

def cython_utility(name, **kwargs):
    return cython_module(PROJECT_NAME, 'utils', 'ext', name, **kwargs)

def additional_source(*args):
    return os.path.join(
           os.path.relpath(BASE_PATH, start=CWD), *args)

# PROJECT VERSION & METADATA
__version__ = "<undefined>"
try:
    exec(compile(
        open(os.path.join(BASE_PATH,
            '__version__.py')).read(),
            '__version__.py', 'exec'))
except:
    print("ERROR COMPILING __version__.py")
    __version__ = '0.8.0'

# PROJECT DESCRIPTION
LONG_DESCRIPTION = project_content('ABOUT.md')

# SOFTWARE LICENSE
LICENSE = 'MIT'

# REQUIRED INSTALLATION DEPENDENCIES
INSTALL_REQUIRES = project_content('requirements', 'install.txt').splitlines()

# PYPI PROJECT CLASSIFIERS
CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'License :: OSI Approved :: MIT License',
    'Intended Audience :: Developers',
    'Operating System :: MacOS',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: OS Independent',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7']

# NUMPY: C-API INCLUDE DIRECTORY
try:
    import numpy
except ImportError:
    class FakeNumpy(object):
        def get_include(self):
            return os.path.curdir
    numpy = FakeNumpy()

# SOURCES & INCLUDE DIRECTORIES
hsluv_source = additional_source('utils', 'ext', 'hsluv.c')
augli_source = additional_source('comparators', 'ext', 'butteraugli.cc')
include_dirs = [numpy.get_include(),
                sysconfig.get_path('include')]

# THE CALL TO `setup(…)`
setup(
    name=PROJECT_NAME,
    author=AUTHOR_NAME,
    author_email=AUTHOR_EMAIL,
    
    version=__version__,
    description=__doc__.strip(),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    
    keywords=" ".join(KEYWORDS),
    url=PROJECT_GH_URL, download_url=PROJECT_DL_URL,
    license=LICENSE, platforms=['any'],
    classifiers=CLASSIFIERS,
    
    packages=find_packages(),
    package_data={ '' : ['*.*'] },
    include_package_data=True,
    zip_safe=False,
    
    install_requires=INSTALL_REQUIRES,
    include_dirs=include_dirs,
    
    ext_modules=cythonize([
        cython_comparator("buttereye",  sources=[augli_source],
                                        language="c++"),
        cython_processor("halftone",    include_dirs=include_dirs,
                                        language="c"),
        cython_utility("api",           sources=[hsluv_source],
                                        language="c")
        ], nthreads=cpu_count(),
           compiler_directives=dict(language_level=3,
                                    infer_types=True,
                                    embedsignature=True)),
)




from PIL import ImageOps, ImageChops
from PIL.ImageEnhance import Color as _Color
from PIL.ImageEnhance import Brightness as _Brightness
from PIL.ImageEnhance import Contrast as _Contrast
from PIL.ImageEnhance import Sharpness as _Sharpness

from instakit.abc import Processor

class Adjustment(Processor):
    
    __slots__ = ('value',)
    
    def __init__(self, value=1.0):
        self.value = value
    
    def adjust(self, image):
        return image
    
    def process(self, image):
        if self.value != 1.0:
            return self.adjust(image)
        return image

class Color(Adjustment):
    """ Globally tweak the image color """
    __slots__ = tuple()
    def adjust(self, image):
        return _Color(image).enhance(self.value)

class Brightness(Adjustment):
    """ Adjust the image brightness """
    __slots__ = tuple()
    def adjust(self, image):
        return _Brightness(image).enhance(self.value)

class Contrast(Adjustment):
    """ Adjust the image contrast """
    __slots__ = tuple()
    def adjust(self, image):
        return _Contrast(image).enhance(self.value)

class Sharpness(Adjustment):
    """ Adjust the sharpness of the image """
    __slots__ = tuple()
    def adjust(self, image):
        return _Sharpness(image).enhance(self.value)

class Invert(Processor):
    __slots__ = tuple()
    def process(self, image):
        return ImageChops.invert(image)

class Equalize(Processor):
    """ Apply a non-linear mapping to the image, via histogram """
    __slots__ = ('mask',)
    
    def __init__(self, mask=None):
        self.mask = hasattr(mask, 'copy') and mask.copy() or mask
    
    def process(self, image):
        return ImageOps.equalize(image, mask=self.mask)

class AutoContrast(Processor):
    """ Normalize contrast throughout the image, via histogram """
    __slots__ = ('cutoff', 'ignore')
    
    def __init__(self, cutoff=0, ignore=None):
        self.cutoff = cutoff
        self.ignore = ignore
    
    def process(self, image):
        return ImageOps.autocontrast(image, cutoff=self.cutoff,
                                            ignore=self.ignore)

class Solarize(Processor):
    """ Invert all pixel values above an 8-bit threshold """
    __slots__ = ('threshold',)
    
    def __init__(self, threshold=128):
        self.threshold = min(max(1, threshold), 255)
    
    def process(self, image):
        return ImageOps.solarize(image, threshold=self.threshold)

class Posterize(Processor):
    """ Reduce the number of bits (1 to 8) per channel """
    __slots__ = ('bits',)
    
    def __init__(self, bits=4):
        self.bits = min(max(1, bits), 8)
    
    def process(self, image):
        return ImageOps.posterize(image, bits=self.bits)
    



from libc.stdint cimport *
from libcpp.vector cimport vector

cdef extern from "butteraugli.h" namespace "butteraugli" nogil:
    
    cppclass Image[ComponentType]:
        Image()
        Image(size_t const, size_t const)       # xsize, ysize
        Image(size_t const, size_t const,       # 
                            ComponentType val)  # xsize, ysize, component-type
        Image(size_t const, size_t const,       # xsize, ysize,
              uint8_t*,     size_t const)       # byteptr, bytes-per-row
        
        Image(Image&&)                          # move constructor
        Image& operator=(Image&&)               # move assignment operator
        
        size_t xsize() const
        size_t ysize() const
        
        const ComponentType* Row(size_t const)
        const uint8_t* byte_ptr "bytes"()
        
        size_t bytes_per_row()
        intptr_t PixelsPerRow() const

ctypedef Image[float]   ImageF
ctypedef Image[uint8_t] Image8

cdef extern from "butteraugli.h" namespace "butteraugli" nogil:
    
    cdef vector[Image[T]] CreatePlanes[T](size_t const,
                                          size_t const,
                                          size_t const)

ctypedef vector[ImageF]     imagefvec
ctypedef vector[Image8]     image8vec
ctypedef vector[float]      floatvec
ctypedef vector[floatvec]   floatvecvec


cdef extern from "butteraugli.h" namespace "butteraugli" nogil:
    
    cdef bint ButteraugliInterface(imagefvec&,
                                   imagefvec&, float,
                                   ImageF&,
                                   double&)
    
    cdef double ButteraugliFuzzyClass(double)
    cdef double ButteraugliFuzzyInverse(double)
    
    cdef bint ButteraugliAdaptiveQuantization(size_t, size_t,
                                              floatvecvec&,
                                              floatvec&)
    
    cdef const double kButteraugliQuantLow
    cdef const double kButteraugliQuantHigh



#!/usr/bin/env python
# encoding: utf-8
"""
noise.py

Created by FI$H 2000 on 2014-05-23.
Copyright (c) 2012-2019 Objects In Space And Time, LLC. All rights reserved.
"""
from __future__ import print_function
from enum import Enum, unique

from instakit.utils.ndarrays import NDProcessor


@unique
class NoiseMode(Enum):
    
    LOCALVAR        = 'localvar'
    GAUSSIAN        = 'gaussian'
    POISSON         = 'poisson'
    SALT            = 'salt'
    PEPPER          = 'pepper'
    SALT_N_PEPPER   = 's&p'
    SPECKLE         = 'speckle'
    
    def to_string(self):
        return str(self.value)
    
    def __str__(self):
        return self.to_string()
    
    def process_nd(self, ndimage, **kwargs):
        from skimage.util import random_noise
        return random_noise(ndimage,
                            mode=self.to_string(),
                          **kwargs)


class Noise(NDProcessor):
    """ Base noise processor
        -- defaults to “localvar” mode; q.v. `GaussianLocalVarianceNoise` sub.
    """
    mode = NoiseMode.LOCALVAR
    
    def process_nd(self, ndimage):
        noisemaker = type(self).mode
        return self.compand(noisemaker.process_nd(ndimage))


class GaussianNoise(Noise):
    """ Add Gaussian noise """
    mode = NoiseMode.GAUSSIAN

class PoissonNoise(Noise):
    """ Add Poisson-distributed noise """
    mode = NoiseMode.POISSON

class GaussianLocalVarianceNoise(Noise):
    """ Add Gaussian noise, with image-dependant local variance """
    pass

class SaltNoise(Noise):
    """ Add “salt noise”
        -- replace random pixel values with 1.0f (255)
    """
    mode = NoiseMode.SALT

class PepperNoise(Noise):
    """ Add “pepper noise”
        -- replace random pixel values with zero
    """
    mode = NoiseMode.PEPPER

class SaltAndPepperNoise(Noise):
    """ Add “salt and pepper noise”
        -- replace random pixel values with either 1.0f (255) or zero
    """
    mode = NoiseMode.SALT_N_PEPPER

class SpeckleNoise(Noise):
    """ Add “speckle noise”
        --- multiplicative noise using `out = image + n * image`
           (where `n` is uniform noise with specified mean + variance)
    """
    mode = NoiseMode.SPECKLE


def test():
    from instakit.utils.static import asset
    from instakit.utils.mode import Mode
    
    image_paths = list(map(
        lambda image_file: asset.path('img', image_file),
            asset.listfiles('img')))
    image_inputs = list(map(
        lambda image_path: Mode.RGB.open(image_path),
            image_paths))
    
    noises = [GaussianNoise,
              PoissonNoise,
              GaussianLocalVarianceNoise,
              SaltNoise,
              PepperNoise,
              SaltAndPepperNoise,
              SpeckleNoise]
    
    for idx, image_input in enumerate(image_inputs + image_inputs[:2]):
        for NoiseProcessor in noises:
            NoiseProcessor().process(image_input).show()
    
    print(image_paths)

if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
from __future__ import print_function

import contextlib
import numpy
import os

from PIL import Image, ImageMode
from enum import auto, unique

from instakit.abc import Enum
from instakit.utils import misc

junkdrawer = misc.Namespace()
junkdrawer.imode = lambda image: ImageMode.getmode(image.mode)

def split_abbreviations(s):
    """ Split a string into a tuple of its unique constituents,
        based on its internal capitalization -- to wit:
        
        >>> split_abbreviations('RGB')
        ('R', 'G', 'B')
        >>> split_abbreviations('CMYK')
        ('C', 'M', 'Y', 'K')
        >>> split_abbreviations('YCbCr')
        ('Y', 'Cb', 'Cr')
        >>> split_abbreviations('sRGB')
        ('R', 'G', 'B')
        >>> split_abbreviations('XYZZ')
        ('X', 'Y', 'Z')
        >>> split_abbreviations('I;16B')
        ('I',)
        
        If you still find this function inscrutable,
        have a look here: https://gist.github.com/4027079
    """
    abbreviations = []
    current_token = ''
    for char in s.split(';')[0]:
        if current_token == '':
            current_token += char
        elif char.islower():
            current_token += char
        else:
            if not current_token.islower():
                if current_token not in abbreviations:
                    abbreviations.append(current_token)
            current_token = ''
            current_token += char
    if current_token != '':
        if current_token not in abbreviations:
            abbreviations.append(current_token)
    return tuple(abbreviations)


ImageMode.getmode('RGB') # one call must be made to getmode()
                         # to properly initialize ImageMode._modes:

junkdrawer.modes = ImageMode._modes
junkdrawer.types = Image._MODE_CONV
junkdrawer.ismap = Image._MAPMODES

mode_strings = tuple(junkdrawer.modes.keys())
dtypes_for_modes = { k : v[0] for k, v in junkdrawer.types.items() }

junkdrawer.idxmode = lambda idx: ImageMode.getmode(mode_strings[idx])
junkdrawer.is_mapped = lambda mode: mode in junkdrawer.ismap

class ModeAncestor(Enum):
    """
    Valid ImageMode mode strings:
    ('1',    'L',     'I',     'F',     'P',
     'RGB',  'RGBX',  'RGBA',  'CMYK',  'YCbCr',
     'LAB',  'HSV',   'RGBa',  'LA',    'La',
     'PA',   'I;16',  'I;16L', 'I;16B')
    """
    
    def _generate_next_value_(name,
                              start,
                              count,
                              last_values):
        return junkdrawer.idxmode(count)
    
    @classmethod
    def _missing_(cls, value):
        try:
            return cls(junkdrawer.idxmode(value))
        except (IndexError, TypeError):
            pass
        return super(ModeAncestor, cls)._missing_(value)
    
    @classmethod
    def is_mode(cls, instance):
        return type(instance) in cls.__mro__

junkdrawer.or_none = lambda thing, name: getattr(thing, name, None)

class ModeContext(contextlib.AbstractContextManager):
    
    """ An ad-hoc mutable named-tuple-ish context-manager class,
        for keeping track of an image while temporarily converting
        it to a specified mode within the managed context block.
        
        Loosely based on the following Code Review posting:
        • https://codereview.stackexchange.com/q/173045/6293
    """
    
    __slots__ = ('initial_image',
                         'image',
                   'final_image',
                 'original_mode',
                          'mode',
                       'verbose')
    
    def __init__(self, image, mode, **kwargs):
        assert Image.isImageType(image)
        assert Mode.is_mode(mode)
        self.verbose = bool(kwargs.get('verbose', False))
        if self.verbose:
            label = junkdrawer.or_none(image, 'filename') \
                and os.path.basename(getattr(image, 'filename')) \
                or str(image)
            print("ModeContext.__init__: configured with image: %s" % label)
        self.initial_image = image
        self.image = None
        self.final_image = None
        self.original_mode = Mode.of(image)
        self.mode = mode
    
    def __repr__(self):
        return type(self).__name__ + repr(tuple(self))
    
    def __iter__(self):
        for name in self.__slots__:
            yield getattr(self, name)
    
    def __getitem__(self, idx):
        return getattr(self, self.__slots__[idx])
    
    def __len__(self):
        return len(self.__slots__)
    
    def attr_or_none(self, name):
        return junkdrawer.or_none(self, name)
    
    def attr_set(self, name, value):
        setattr(self, name, value)
    
    def __enter__(self):
        initial_image = self.attr_or_none('initial_image')
        mode = self.attr_or_none('mode')
        if initial_image is not None and mode is not None:
            if self.verbose:
                print("ModeContext.__enter__: converting %s to %s" % (
                       Mode.of(initial_image),
                       mode))
            image = mode.process(initial_image)
            self.attr_set('image', image)
        return self
    
    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        image = self.attr_or_none('image')
        original_mode = self.attr_or_none('original_mode')
        if image is not None and original_mode is not None:
            if self.verbose:
                print("ModeContext.__exit__: converting %s to %s" % (
                       Mode.of(image),
                       original_mode))
            final_image = original_mode.process(image)
            self.attr_set('final_image', final_image)
        return exc_type is None


@unique
class Mode(ModeAncestor):
    
    """ An enumeration class wrapping ImageMode.ModeDescriptor. """
    
    # N.B. this'll have to be manually updated,
    # whenever PIL.ImageMode gets a change pushed.
    
    MONO    = auto() # formerly ‘1’
    L       = auto()
    I       = auto()
    F       = auto()
    P       = auto()
    
    RGB     = auto()
    RGBX    = auto()
    RGBA    = auto()
    CMYK    = auto()
    YCbCr   = auto()
    
    LAB     = auto()
    HSV     = auto()
    RGBa    = auto()
    LA      = auto()
    La      = auto()
    
    PA      = auto()
    I16     = auto() # formerly ‘I;16’
    I16L    = auto() # formerly ‘I;16L’
    I16B    = auto() # formerly ‘I;16B’
    
    @classmethod
    def of(cls, image):
        for mode in cls:
            if mode.check(image):
                return mode
        raise ValueError("Image has unknown mode %s" % image.mode)
    
    @classmethod
    def for_string(cls, string):
        for mode in cls:
            if mode.to_string() == string:
                return mode
        raise ValueError("for_string(): unknown mode %s" % string)
    
    def to_string(self):
        return str(self.value)
    
    def __str__(self):
        return self.to_string()
    
    def __repr__(self):
        repr_string = "%s(%s: [%s/%s] ∞ {%s » %s}) @ %s"
        return repr_string % (type(self).__qualname__,
                                   self.label,
                                   self.basemode, self.basetype,
                                   self.dtype_code(), self.dtype,
                                id(self))
    
    def __bytes__(self):
        return bytes(self.to_string(), encoding="UTF-8")
    
    def __call__(self, image, **kwargs):
        return ModeContext(image, self, **kwargs)
    
    def dtype_code(self):
        return dtypes_for_modes.get(self.to_string(), None) or \
                                    self.basetype.dtype_code()
    
    @property
    def band_count(self):
        return len(self.value.bands)
    
    @property
    def bands(self):
        return self.value.bands
        
    @property
    def basemode(self):
        return type(self).for_string(self.value.basemode)
    
    @property
    def basetype(self):
        return type(self).for_string(self.value.basetype)
    
    @property
    def dtype(self):
        return numpy.dtype(self.dtype_code())
    
    @property
    def is_memory_mapped(self):
        return junkdrawer.is_mapped(self.to_string())
    
    @property
    def label(self):
        return str(self) == self.name \
                        and self.name \
                        or "%s (%s)" % (self, self.name)
    
    def check(self, image):
        return junkdrawer.imode(image) is self.value
    
    def merge(self, *channels):
        return Image.merge(self.to_string(), channels)
    
    def render(self, image, *args, **kwargs):
        if self.check(image):
            return image
        return image.convert(self.to_string(),
                            *args,
                           **kwargs)
    
    def process(self, image):
        return self.render(image)
    
    def new(self, size, color=0):
        return Image.new(self.to_string(), size, color=color)
    
    def open(self, fileish):
        return self.render(Image.open(fileish))
    
    def frombytes(self, size, data, decoder_name='raw', *args):
        return Image.frombytes(self.to_string(),
                               size, data, decoder_name,
                              *args)


def test():
    from pprint import pprint
    
    print("«KNOWN IMAGE MODES»")
    print()
    
    for m in Mode:
        print("• %10s\t ∞%5s/%s : %s » %s" % (m.label,
                                              m.basemode,
                                              m.basetype,
                                              m.dtype_code(),
                                              m.dtype))
    
    print()
    
    """
    •   1 (MONO)	 ∞    L/L : |b1 » bool
    •          L	 ∞    L/L : |u1 » uint8
    •          I	 ∞    L/I : <i4 » int32
    •          F	 ∞    L/F : <f4 » float32
    •          P	 ∞  RGB/L : |u1 » uint8
    •        RGB	 ∞  RGB/L : |u1 » uint8
    •       RGBX	 ∞  RGB/L : |u1 » uint8
    •       RGBA	 ∞  RGB/L : |u1 » uint8
    •       CMYK	 ∞  RGB/L : |u1 » uint8
    •      YCbCr	 ∞  RGB/L : |u1 » uint8
    •        LAB	 ∞  RGB/L : |u1 » uint8
    •        HSV	 ∞  RGB/L : |u1 » uint8
    •       RGBa	 ∞  RGB/L : |u1 » uint8
    •         LA	 ∞    L/L : |u1 » uint8
    •         La	 ∞    L/L : |u1 » uint8
    •         PA	 ∞  RGB/L : |u1 » uint8
    • I;16 (I16)	 ∞    L/L : <u2 » uint16
    • I;16L (I16L)	 ∞    L/L : <u2 » uint16
    • I;16B (I16B)	 ∞    L/L : >u2 » >u2
    
    """
    
    print("«TESTING: split_abbreviations()»")
    
    assert split_abbreviations('RGB') == ('R', 'G', 'B')
    assert split_abbreviations('CMYK') == ('C', 'M', 'Y', 'K')
    assert split_abbreviations('YCbCr') == ('Y', 'Cb', 'Cr')
    assert split_abbreviations('sRGB') == ('R', 'G', 'B')
    assert split_abbreviations('XYZZ') == ('X', 'Y', 'Z')
    assert split_abbreviations('I;16L') == ('I',)
    
    assert split_abbreviations('RGB') == Mode.RGB.bands
    assert split_abbreviations('CMYK') == Mode.CMYK.bands
    assert split_abbreviations('YCbCr') == Mode.YCbCr.bands
    assert split_abbreviations('I;16L') == Mode.I16L.bands
    assert split_abbreviations('sRGB') == Mode.RGB.bands
    # assert split_abbreviations('XYZ') == ('X', 'Y', 'Z')
    
    print("«SUCCESS»")
    print()
    
    # print(Mode.I16L.bands)
    # print(Mode.RGB.bands)
    
    pprint(list(Mode))
    print()
    
    assert Mode(10) == Mode.LAB
    assert hasattr(Mode.RGB, '__slots__')
    
    print()
    
    print("«TESTING: CONTEXT-MANAGED IMAGE MODES»")
    print()
    from instakit.utils.static import asset
    
    image_paths = list(map(
        lambda image_file: asset.path('img', image_file),
            asset.listfiles('img')))
    image_inputs = list(map(
        lambda image_path: Mode.RGB.open(image_path),
            image_paths))
    
    for image in image_inputs:
        with Mode.L(image, verbose=True) as grayscale:
            assert Mode.of(grayscale.image) is Mode.L
            print(grayscale.image)
            grayscale.image = Mode.MONO.process(grayscale.image)
        print()
    
    print("«SUCCESS»")
    print()

if __name__ == '__main__':
    test()



#!/usr/bin/env python
# encoding: utf-8
"""
lutmap.py

Created by FI$H 2000 on 2012-08-23.
Copyright (c) 2012 Objects In Space And Time, LLC. All rights reserved.
"""
from __future__ import print_function

import numpy, imread
from os.path import join
from collections import defaultdict

from PIL import Image
#from math import floor

from instakit.utils.colortype import ColorType
from instakit.utils import static

class RGBTable(defaultdict):
    RGB = ColorType('RGB', dtype=numpy.uint8)
    identity = numpy.zeros(
        shape=(512, 512),
        dtype=numpy.uint64)
    
    for bx in range(0, 8):
        for by in range(0, 8):
            for r in range(0, 64):
                for g in range(0, 64):
                    identity[
                        int(g + float(by) * 64.0 + 0.5),
                        int(r + float(bx) * 64.0 + 0.5)] = hash(RGB(
                            int(r * 255.0 / 63.0 + 0.5),
                            int(g * 255.0 / 63.0 + 0.5),
                            int((bx + by * 8.0) * 255.0 / 63.0 + 0.5)))
    
    def __init__(self):
        super(RGBTable, self).__init__(default_factory=None)
        self.data = self.identity
    
    def __missing__(self, color):
        self[color] = value = self.lookup(color)
        return value
    
    def _idx(self, color):
        print("_idx COLOR:")
        print(color)
        print("_idx WAT:")
        print(int('%02x%02x%02x' % color, 16))
        print("hash COLOR:")
        print(hash(color))
        return int('%02x%02x%02x' % color, 16)
    
    def _rgb(self, idx):
        RGB = self.RGB
        return RGB(*reversed(
            [(idx >> (8*i)) & 255 for i in range(3)]))
    
    def lookup(self, color):
        print("lookup COLOR:")
        print(color)
        return self.color_at(*self._xy(color))
    
    def _xy(self, color):
        where = numpy.where(
            self.identity[:,:] == hash(color))
        print("WHERE:")
        print(len(tuple(zip(*where))))
        try:
            return tuple(zip(*where))[0]
        except IndexError:
            return []
    
    def color_at(self, x, y, data=None):
        print("X, Y: %s, %s" % (x, y))
        print("data: %s" % data)
        if data is None:
            data = self.data
        print("DATA.shape:")
        print(self.data.shape)
        print(data[x, y])
        return self.RGB(*data[x, y])
    
    def float_color_at(self, x, y, data=None):
        if data is None:
            data = self.identity
        return (channel/255.0 for channel in self.color_at(x, y, data=data))


class LUT(RGBTable):
    
    def __init__(self, name='identity'):
        RGBTable.__init__(self)
        self.name = name
        self.data = self._read_png_matrix(self.name)
    
    @classmethod
    def _read_png_matrix(cls, name):
        print("Reading LUT image: %s" % static.path(join('lut', '%s.png' % name)))
        return imread.imread(
            static.path(join('lut', '%s.png' % name)))


def main():

    RGB = ColorType('RGB')
    RGB24 = ColorType('RGB', dtype=numpy.uint8)
    YCrCb = ColorType('YCrCb', dtype=numpy.uint8)
    
    print(RGB(2, 3, 4))
    print(RGB24)
    print(YCrCb(8, 88, 808))
    
    identity = LUT()
    amatorka = LUT('amatorka')
    
    print(identity.identity)
    print(RGB(22,33,44))
    print(int(RGB(22,33,44)))
    print(int(RGB(55,66,77)))
    print(numpy.any(identity.identity[:,:] == int(RGB(11,44,99))))
    print(numpy.max(identity.identity))
    print(RGB(111, 222, 11).composite)
    
    print("")
    print(identity[RGB(146,146,36)])
    #print(identity[RGB(22,33,44)])
    print(identity[RGB(132, 166, 188)])
    
    print("")
    print("YO DOGG")
    print(amatorka[RGB(146,146,36)])
    print(identity[RGB(22,33,44)])
    print(identity[RGB(255, 25, 25)])
    

def blurthday():
    
    from imread import imread
    from pprint import pprint
    imfuckingshowalready = lambda mx: Image.fromarray(mx).show()
    
    identity = LUT()
    amatorka = LUT('amatorka')
    #miss_etikate = LUT('miss_etikate')
    #soft_elegance_1 = LUT('soft_elegance_1')
    #soft_elegance_2 = LUT('soft_elegance_2')
    
    im1 = imread(static.path(join('img', '06-DSCN4771.JPG')))
    im2 = imread(static.path(join(
        'img', '430023_3625646599363_1219964362_3676052_834528487_n.jpg')))
    
    pprint(identity)
    pprint(amatorka)
    
    im9 = amatorka.transform(im1)
    pprint(im9)
    imfuckingshowalready(im9)
    print(im1)
    print(im2)


def old_maid():
    pass
    #global __multipons__
    #from pprint import pprint
    #pprint(__multipons__)

def old_main():
    
    #imfuckingshowalready = lambda mx: Image.fromarray(mx).show()

    old_identity = static.path(join('lut', 'identity.png'))

    im_old_identity = imread.imread(old_identity)
    im_identity = numpy.zeros_like(im_old_identity)

    for bx in range(0, 8):
        for by in range(0, 8):
            for r in range(0, 64):
                for g in range(0, 64):
                    im_identity[
                        int(g + by * 64),
                        int(r + bx * 64)] = numpy.array((
                            int(r * 255.0 / 63.0 + 0.5),
                            int(g * 255.0 / 63.0 + 0.5),
                                int((bx + by * 8.0) * 255.0 / 63.0 + 0.5)),
                                dtype=numpy.uint8)
    
    print("THE OLD: %s, %s, %s" % (
        im_old_identity.size, im_old_identity.shape,
        str(im_old_identity.dtype)))
    #print(im_old_identity)
    print("")
    
    print("THE NEW: %s, %s, %s" % (
        im_identity.size, im_identity.shape,
        str(im_identity.dtype)))
    #print(im_identity)
    print("")
    
    
    
    print("THE END: %s" % bool(im_old_identity.shape == im_identity.shape))
    #print(im_old_identity == im_identity)
    
    #imfuckingshowalready(im_identity)
    #imfuckingshowalready(im_old_identity)
    
    pil_im_old_identity = Image.fromarray(im_old_identity)
    pil_im_old_identity.save('/tmp/im_old_identity.jpg',
        format="JPEG")
    
    pil_im_identity = Image.fromarray(im_identity)
    pil_im_identity.save('/tmp/im_identity.jpg',
        format="JPEG")




if __name__ == '__main__':
    main()






#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function
import os

projectdir = os.path.join(os.path.dirname(__file__), '..', '..')
namespaces = set()

def static_namespace(name):
    """ Configure and return an instakit.utils.misc.Namespace instance,
        festooning it with shortcuts allowing for accesing static files
        within subdirectories of the Instakit project package tree.
    """
    from instakit.utils.misc import Namespace
    ns = Namespace()
    ns.name = str(name)
    ns.root = os.path.join(projectdir, ns.name)
    ns.data = os.path.join(ns.root, 'data')
    ns.relative = lambda p: os.path.relpath(p, start=ns.root)
    ns.listfiles = lambda *p: os.listdir(os.path.join(ns.data, *p))
    ns.path = lambda *p: os.path.abspath(os.path.join(ns.data, *p))
    namespaces.add(ns)
    return ns

asset = static_namespace('instakit')
tests = static_namespace('tests')

__all__ = ('projectdir',
           'namespaces',
           'static_namespace',
           'asset', 'tests')

__dir__ = lambda: list(__all__)

def test():
    assert os.path.isdir(projectdir)
    assert len(namespaces) == 2
    
    assert os.path.isdir(asset.root)
    assert os.path.isdir(asset.data)
    assert len(asset.listfiles('acv')) > 0
    assert len(asset.listfiles('icc')) > 0
    assert len(asset.listfiles('img')) > 0
    assert len(asset.listfiles('lut')) > 0
    
    assert os.path.isdir(tests.root)
    assert os.path.isdir(tests.data)
    assert len(os.listdir(tests.data)) > 0

if __name__ == '__main__':
    test()
